import logging
from stanag4586edav1.message_wrapper import *
from stanag4586edav1.message01 import *
from stanag4586edav1.message20 import *
from stanag4586edav1.message21 import *
from stanag4586edav1.message200 import *
from stanag4586edav1.message300 import *
from stanag4586edav1.message301 import *
from stanag4586edav1.message302 import *
from stanag4586edav1.message1200 import *
from stanag4586edav1.message20000 import *
from stanag4586edav1.message20010 import *
from stanag4586edav1.message20020 import *
class StanagProtocol:

    def __init__(self, loop, debug_level, on_msg_rx_callback, on_con_lost_callback, rx_enabled = True):
        
        self.loop = loop
        self.transport = None

        self.on_msg_rx_callback = on_msg_rx_callback
        self.on_con_lost_callback = on_con_lost_callback
        
        self.logger = logging.getLogger('StanagProtocol')
        self.logger.setLevel(debug_level)

        self.rx_enabled = rx_enabled

    def connection_made(self, transport):
        self.transport = transport
    
    def error_received(self, exc):
        self.logger.error('Error received [{}]'.format(exc))

    def connection_lost(self, exc):
        self.logger.error("Connection lost")
        self.loop.call_soon(self.on_con_lost_callback)
        
    def datagram_received(self, data, addr):

        if not self.rx_enabled:
            self.logger.warn("Rx is disabled and yet got a message on this socket.")
            return

        self.logger.debug("Got packet of len [{}]".format(len(data)))

        wrapper = MessageWrapper(data)
        self.logger.debug("Got message [{:}]".format(wrapper.message_type))

        msg = None

        known_messages = {
            1 : Message01,
            20 : Message20,
            21 : Message21,
            200 : Message200,
            300 : Message300,
            301 : Message301,
            302 : Message302,
            1200 : Message1200,
            20000 : Message20000,
            20010 : Message20010,
            20020 : Message20020,
        }
        

        if wrapper.message_type in known_messages.keys():
            msg_type_to_instantiate = known_messages[wrapper.message_type]
            msg = msg_type_to_instantiate(data[MessageWrapper.MSGLEN:])

        if msg is not None:
            self.logger.debug("callback scheduled")
            self.loop.call_soon(self.on_msg_rx_callback, wrapper, msg)



import asyncio
import socket
import struct
from .stanag_protocol import StanagProtocol
from .controllable_entity import ControllableEntity
import logging
from stanag4586edav1.message300 import *

STANAG_SERVER_MODE_VEHICLE = 0
STANAG_SERVER_MODE_CUCS = 1

class StanagServer:

    MODE_VEHICLE = 0
    MODE_CUCS = 1
    
    __controllable_entities = {}
    __mode = MODE_VEHICLE
    
    def __init__(self, debug_level):
        self.logger = logging.getLogger('StanagServer')
        self.logger.setLevel(debug_level)
        self.debug_level = debug_level

    async def setup_service(self, loop, mode, port_rx = 4586, port_tx = 4587, addr_rx = "224.10.10.10", addr_tx = "224.10.10.10"):

        self.logger.info("Server setup Started.")

        self.__mode == mode

        if mode is self.MODE_VEHICLE:
            self.logger.info("Creating entities.")
            self.create_entities(loop)

        self.logger.info("Setting up network i/o.")

        if mode is self.MODE_VEHICLE:
            await self.create_rx_socket(loop, port_rx, addr_rx)
            await self.create_tx_socket(loop, port_tx, addr_tx)
        else:
            await self.create_rx_socket(loop, port_tx, addr_tx)
            await self.create_tx_socket(loop, port_rx, addr_rx)

        self.logger.info("Server setup completed.")

    def on_rx_con_lost(self):
        pass

    async def create_rx_socket(self, loop, port_rx, addr_rx):
        self.__sock_rx = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        self.__sock_rx.bind(('', port_rx))
        group = socket.inet_aton(addr_rx)
        mreq = struct.pack('4sL', group, socket.INADDR_ANY)
        self.__sock_rx.setsockopt(socket.IPPROTO_IP, socket.IP_ADD_MEMBERSHIP, mreq)
        self.__sock_rx.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        self.__sock_rx.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEPORT, 1)

        self.__transport_rx, self.__protocol_rx = await loop.create_datagram_endpoint(
            lambda: StanagProtocol(loop, self.debug_level, self.on_msg_rx, self.on_rx_con_lost, True),
            sock=self.__sock_rx,
        )
    
    def on_tx_con_lost(self):
        pass

    async def create_tx_socket(self, loop, port_tx, addr_tx):

        self.__transport_tx, self.__protocol_tx = await loop.create_datagram_endpoint(
            lambda: StanagProtocol(loop, self.debug_level, self.on_msg_rx, self.on_tx_con_lost, False),
            remote_addr=(addr_tx, port_tx),
        )

    def tx_data(self, data):
        """Transmit the data on network using __transport_tx, returns nothing."""
        if self.__transport_tx is not None:
            self.__transport_tx.sendto(data)
        
    def create_entities(self, loop):
        """Creates all the sensors and stations that can be controlled by CUCS. Returns nothing."""
        
        self.__controllable_entities['base'] = ControllableEntity(loop, self.debug_level, 0x0, 0x1, 0x1, self.tx_data)
        self.__controllable_entities['eo'] = ControllableEntity(loop, self.debug_level, 0x1, 0x1, 0x1, self.tx_data)
        
        self.__controllable_entities['base'].set_available_stations(0x01)
        self.__controllable_entities['eo'].set_payload_type(Message300.STATION_TYPE_EO)

    def get_entity(self, entity_name):
        if entity_name in self.__controllable_entities.keys():
            return self.__controllable_entities[entity_name]
                
    def on_msg_rx(self, wrapper, msg):
        """Callback passed to stanag protocal and is invoked when a known message arrives."""
        self.logger.debug("Got message [{}]".format(wrapper.message_type))

        if self.__mode is self.MODE_VEHICLE:
            """If running on the vehicle end"""
            for k,v in self.__controllable_entities.items():
                if True == v.handle_message(wrapper, msg):
                    return

            if wrapper.message_type != 1:
                # got a message that was not handled....
                self.logger.warn('Message [{}] not handled'.format(wrapper.message_type))

        elif self.__mode is self.MODE_CUCS:
            """If running on the cucs side"""
            pass




"""
 Copyright (c) 2021 Faisal Thaheem (https://github.com/faisalthaheem/python-stanag-4586-EDA-v1)
 License GNU GENERAL PUBLIC LICENSE Version 3, 29 June 2007
"""

import pathlib
from setuptools import find_packages, setup

HERE = pathlib.Path(__file__).parent
README = (HERE / "README.md").read_text()

setup(
    name='stanag4586vsm',
    packages=find_packages(include=['stanag4586vsm']),
    version='0.0.4',
    description='Python Stanag 4586 VSM implementation',
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://github.com/faisalthaheem/python-stanag-4586-vsm",
    author='Faisal Thaheem',
    license='GPLV3',
    classifiers=[
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Programming Language :: Python :: 3",
    ],
    install_requires=[],
    setup_requires=['pytest-runner'],
    tests_require=['pytest'],
    test_suite='tests',
)


from enum import auto
import logging
import pprint
from stanag4586edav1.message_wrapper import *
from stanag4586edav1.message01 import *
from stanag4586edav1.message20 import *
from stanag4586edav1.message21 import *
from stanag4586edav1.message300 import *

class ControllableEntity:

    __station_id = 0x0 #by default we are the base platform
    __vsm_id = 0x0
    __vehicle_id = 0x0
    __monitoring_cucs_list = []
    __controlling_cucs_id = 0x0
    __loop = None
    __available_stations = 0x00
    __payload_type = 0x00
    __callback_unhandled_messages = None

    def __init__(self, loop, debug_level, station_id, vsm_id, vehicle_id, callback_tx_data):
        self.__loop = loop
        self.__station_id = station_id
        self.__vsm_id = vsm_id
        self.__vehicle_id = vehicle_id
        self.__callback_tx_data = callback_tx_data

        self.logger = logging.getLogger('ControllableEntity[{}]'.format(self.__station_id))
        self.logger.setLevel(debug_level)

    # returns true if the message is handled
    def handle_message(self, wrapper, msg):
        self.logger.debug("Got message [{}]".format(wrapper.message_type))

        if wrapper.message_type == 1:
            self.logger.debug("Message is auth request.")
            return self.handle_auth_message(wrapper, msg)

        # filter out unnecessary invocations of outer message handler below by ensuring this entity is the intended recepient
        # of the message
        if msg.has_station_number_field and (msg.station_number != self.__station_id):
            return False
        
        # for messages that do not have station_number we have no choice but to invoke handler
        self.process_incoming_message(wrapper, msg)

        return True

    #for any messages we cannot process in this class
    def set_callback_for_unhandled_messages(self, callback):
        if callback is not None:
            self.__callback_unhandled_messages = callback

    #invokes the __callback_unhandled_messages if it's not None 
    def process_incoming_message(self, wrapper, msg):
        if self.__callback_unhandled_messages is not None:
            try:
                self.logger.debug("Inovking unhandled message handler [{}]".format(wrapper.message_type))
                self.__loop.call_soon(self.__callback_unhandled_messages, wrapper, msg)
            except:
                self.logger.error("Unhandled error: [{}]".format(sys.exc_info()[0]))

    def handle_auth_message(self, wrapper, msg):
    
        if  (msg.vehicle_id & Message01.BROADCAST_ID) == Message01.BROADCAST_ID and \
            (msg.vsm_id & Message01.BROADCAST_ID) == Message01.BROADCAST_ID and \
            (msg.controlled_station & Message01.BROADCAST_ID) == Message01.BROADCAST_ID:
            self.logger.debug("Message is of type discovery")
            self.handle_broadcast(wrapper, msg)
            return False # to let the chain continue to be called at caller end

        elif msg.vehicle_id == self.__vehicle_id:
            self.logger.debug("Message is of type LOI request")
            self.handle_loi_request(wrapper, msg)
            return True

        return False

    def handle_broadcast(self, wrapper, msg):
        
        self.logger.debug("Handling discovery request")

        self.respond_21(wrapper, msg)

        if self.__station_id == 0x0: #base platform
            self.respond_20(wrapper, msg)

        self.respond_300(wrapper, msg)

    def respond_20(self, wrapper, msg):
        self.logger.debug("Responding with Message 20")

        msg20 = Message20(Message20.MSGNULL)

        msg20.time_stamp = 0x00
        msg20.vehicle_id = self.__vehicle_id
        msg20.cucs_id = msg.cucs_id
        msg20.vsm_id = self.__vsm_id
        msg20.vehicle_id_update = 0x0
        msg20.vehicle_type = 0x00 #todo be filled from some config in future
        msg20.vehicle_sub_type = 00 #todo be filled from some config in future
        msg20.owning_id = 0x00 #todo be filled from some config in future
        msg20.set_tail_number("1234") #todo be filled from some config in future
        msg20.set_mission_id("1234") #todo be filled from some config mission
        msg20.set_atc_call_sign("1234") #todo be filled from some config in future
        msg20.configuration_checksum = 0xABCD

        wrapped_reply = MessageWrapper(MessageWrapper.MSGNULL)
        wrapped_reply = wrapped_reply.wrap_message(wrapper.msg_instance_id, 20, msg20, False)

        self.__loop.call_soon(self.__callback_tx_data, wrapped_reply)


    def respond_21(self, wrapper, msg):
        self.logger.debug("Responding with Message 21")
        
        msg21 = Message21(Message21.MSGNULL)

        msg21.time_stamp = 0x00
        msg21.vehicle_id = self.__vehicle_id
        msg21.cucs_id = msg.cucs_id
        msg21.vsm_id = self.__vsm_id
        msg21.data_link_id = 0x0
        msg21.loi_authorized = self.get_loi_authorized(msg.cucs_id)
        msg21.loi_granted = self.get_loi_granted(msg.cucs_id)
        msg21.controlled_station = self.__station_id
        msg21.controlled_station_mode = 1 if ( (msg21.loi_granted & Message01.LOI_05) == Message01.LOI_05) else 0
        msg21.vehicle_type = 0x00 #todo get from config
        msg21.vehicle_sub_type = 0x00 #todo get from config

        wrapped_reply = MessageWrapper(MessageWrapper.MSGNULL)
        wrapped_reply = wrapped_reply.wrap_message(wrapper.msg_instance_id, 21, msg21, False)

        self.__loop.call_soon(self.__callback_tx_data, wrapped_reply)

    def respond_300(self, wrapper, msg):

        """No 300 for BP"""
        if self.__station_id == 0: return

        self.logger.debug("Responding with Message 300")
        
        msg300 = Message300(Message300.MSGNULL)
        msg300.time_stamp = 0x00
        msg300.vehicle_id = self.__vehicle_id
        msg300.cucs_id = msg.cucs_id
        msg300.vsm_id = self.__vsm_id
        msg300.station_number = self.__station_id

        if self.__station_id == 0x0:
            msg300.payload_stations_available = self.__available_stations
            msg300.payload_type = Message300.STATION_TYPE_UNSPECIFIED
        else:
            msg300.payload_stations_available = 0x00
            msg300.payload_type = self.__payload_type
        
        msg300.station_door = 0x00
        msg300.number_of_payload_recording_devices = 0x00

        wrapped_reply = MessageWrapper(MessageWrapper.MSGNULL)
        wrapped_reply = wrapped_reply.wrap_message(wrapper.msg_instance_id, 300, msg300, False)

        self.__loop.call_soon(self.__callback_tx_data, wrapped_reply)

    def get_loi_granted(self, requesting_cucs_id):
        """Calculates and returns the loi_granted field value given a cucs_id"""
        
        granted_loi = 0
        
        if self.__controlling_cucs_id == requesting_cucs_id:
            granted_loi = Message01.LOI_05
        
        if requesting_cucs_id in self.__monitoring_cucs_list:
            """This is a repeat in case the previous statment is true since there is no control without monitoring"""
            granted_loi = (granted_loi or Message01.LOI_02)

        return granted_loi

    def get_loi_authorized(self, requesting_cucs_id):
        """Calculates and returns the loi_authorized field value given a cucs_id"""
        
        """By default everyone can monitor"""
        authorized_loi = Message01.LOI_02

        if self.__controlling_cucs_id in [0, requesting_cucs_id]:
            authorized_loi = (authorized_loi or Message01.LOI_05)
        
        return authorized_loi

    def set_available_stations(self, available_stations):
        self.__available_stations = available_stations
    
    def set_payload_type(self, __payload_type):
        self.__payload_type = __payload_type

    def is_control_bit_set(self, msg):

        CONTROL_LOI_BIT = Message01.LOI_03 if self.__station_id != 0 else Message01.LOI_05
        
        return (msg.requested_handover_loi & CONTROL_LOI_BIT) == CONTROL_LOI_BIT

    def is_monitor_bit_set(self, msg):

        return (msg.requested_handover_loi & Message01.LOI_02) == Message01.LOI_02

    def add_cucs_to_monitoring_list(self, msg):
        
        if msg.cucs_id not in self.__monitoring_cucs_list:
            self.__monitoring_cucs_list.append(msg.cucs_id)
            self.logger.debug("Cucs [{}] added to monitoring list.".format(msg.cucs_id))
        else:
            self.logger.debug("cucs_id already in monitoring list")

        self.logger.debug("Monitoring list is: [{}]".format(self.__monitoring_cucs_list))

    def remove_cucs_from_monitoring_list(self, msg):
        
        if msg.cucs_id in self.__monitoring_cucs_list:
            self.__monitoring_cucs_list.remove(msg.cucs_id)
            self.logger.debug("Cucs [{}] removed from monitoring list.".format(msg.cucs_id))
        else:
            self.logger.debug("cucs_id not present in monitoring list")

        self.logger.debug("Monitoring list is: [{}]".format(self.__monitoring_cucs_list))

    def handle_loi_request(self, wrapper, msg):
        self.logger.debug("Processing LOI request")
        
        if msg.controlled_station_mode == 0x01:
            """asking for something"""

            if self.is_monitor_bit_set(msg):
                """this is a request for monitor"""
                self.add_cucs_to_monitoring_list(msg)

            if self.is_control_bit_set(msg):
                if self.__controlling_cucs_id == 0:
                    self.__controlling_cucs_id = msg.cucs_id
                    self.logger.debug("Control granted to [{}]".format(msg.cucs_id))
                else:
                    self.logger.debug("Cannot grant control to [{}] as already controlled by [{}]".format(msg.cucs_id, self.__controlling_cucs_id))
                
        else:
            """letting go of something"""

            if self.is_monitor_bit_set(msg):
                """this is a request about monitor"""
                self.remove_cucs_from_monitoring_list(msg)

            if self.is_control_bit_set(msg):
                if self.__controlling_cucs_id == msg.cucs_id:
                    self.__controlling_cucs_id = 0
                    self.logger.debug("Control revoked from [{}]".format(msg.cucs_id))
                elif self.__controlling_cucs_id != 0:
                    self.logger.debug("Cannot remove control from [{}] as being controlled by [{}]".format(msg.cucs_id, self.__controlling_cucs_id))

        self.respond_21(wrapper, msg)

